// Copyright 2022 ESRI
//
// All rights reserved under the copyright laws of the United States
// and applicable international laws, treaties, and conventions.
//
// You may freely redistribute and use this sample code, with or
// without modification, provided you include the original copyright
// notice and use restrictions.
//
// See the Sample code usage restrictions document for further information.
//

#include "Display_a_map.h"
#include "Map.h"
#include "MapTypes.h"
#include "MapQuickView.h"
#include "Point.h"
#include "SpatialReference.h"
#include <QFuture>
#include "Viewpoint.h"

#include "FeatureLayer.h"
#include "ServiceFeatureTable.h"
#include "LayerListModel.h"

#include "SymbolTypes.h"
#include "UniqueValue.h"
#include <QUrl>
#include "SimpleFillSymbol.h"
#include "UniqueValueRenderer.h"

#include "SimpleLineSymbol.h"
#include <QList>
#include "ClassBreaksRenderer.h"

#include "SimpleRenderer.h"

#include "LabelDefinition.h"
#include "TextSymbol.h"
#include "ArcadeLabelExpression.h"
#include "LabelDefinitionListModel.h"
#include "PictureMarkerSymbol.h"

using namespace Esri::ArcGISRuntime;

Display_a_map::Display_a_map(QObject* parent /* = nullptr */):
    QObject(parent),
    m_map(new Map(BasemapStyle::ArcGISTopographic, this))
{
}

Display_a_map::~Display_a_map()
{
}

MapQuickView* Display_a_map::mapView() const
{
    return m_mapView;
}

QUrl trailsQUrl("https://services3.arcgis.com/GVgbJbqm8hXASVYi/arcgis/rest/services/Trails/FeatureServer/0");

void Display_a_map::setupViewpoint()
{
    const Point center(-118.80543, 34.02700, SpatialReference::wgs84());
    const Viewpoint viewpoint(center, 100000.0);
    m_mapView->setViewpointAsync(viewpoint);

    addOpenSpaceLayer();

    addTrailsLayer();

    addBikeOnlyTrailsLayer();
    addNoBikeTrailsLayer();

    addTrailheadsLayer();

}

// Create a new function that creates a feature layer and appends it to the operational layers and the m_map.
FeatureLayer* Display_a_map::addFeatureLayer(const QUrl& url)
{
    ServiceFeatureTable* serviceFeatureTable = new ServiceFeatureTable(url, this);
    FeatureLayer* featureLayer = new FeatureLayer(serviceFeatureTable, this);
    m_map->operationalLayers()->append(featureLayer);
    return featureLayer;
}

void Display_a_map::addOpenSpaceLayer()
{
    // Create a parks and open spaces feature layer.
    QUrl parksAndOpenSpacesQUrl("https://services3.arcgis.com/GVgbJbqm8hXASVYi/ArcGIS/rest/services/Parks_and_Open_Space/FeatureServer/0");
    FeatureLayer* featureLayer = addFeatureLayer(parksAndOpenSpacesQUrl);

    // Create fill symbols.
    SimpleFillSymbol* purpleFillSymbol = new SimpleFillSymbol(SimpleFillSymbolStyle::Solid, QColor("purple"), this);
    SimpleFillSymbol* greenFillSymbol = new SimpleFillSymbol(SimpleFillSymbolStyle::Solid, QColor("green"), this);
    SimpleFillSymbol* blueFillSymbol = new SimpleFillSymbol(SimpleFillSymbolStyle::Solid, QColor("blue"), this);
    SimpleFillSymbol* redFillSymbol = new SimpleFillSymbol(SimpleFillSymbolStyle::Solid, QColor("red"), this);

    // Create a unique value for natural areas, regional open spaces, local parks & regional recreation parks.
    UniqueValue* naturalAreas = new UniqueValue("Natural Areas", "Natural Areas", {"Natural Areas"}, purpleFillSymbol, this);
    UniqueValue* regionalOpenSpace = new UniqueValue("Regional Open Space", "Regional Open Space", {"Regional Open Space"}, greenFillSymbol, this);
    UniqueValue* localPark = new UniqueValue("Local Park", "Local Park", {"Local Park"}, blueFillSymbol, this);
    UniqueValue* regionalRecreationPark = new UniqueValue("Regional Recreation Park", "Regional Recreation Park",
                                                          {"Regional Recreation Park"}, redFillSymbol, this);

    // Create and assign a unique value renderer to the feature layer.
    UniqueValueRenderer* openSpacesUniqueValueRenderer = new UniqueValueRenderer("Open Spaces", nullptr, {"TYPE"},
                                                                                 {naturalAreas, regionalOpenSpace, localPark, regionalRecreationPark }, this);
    featureLayer->setRenderer(openSpacesUniqueValueRenderer);

    // Set the layer opacity to semi-transparent.
    featureLayer->setOpacity(0.25f);
}

void Display_a_map::addTrailsLayer()
{
    FeatureLayer* featureLayer = addFeatureLayer(trailsQUrl);
    // Create five line symbols to display class breaks.
    SimpleLineSymbol* firstClassSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("purple"), 3, this);
    SimpleLineSymbol* secondClassSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("purple"), 4, this);
    SimpleLineSymbol* thirdClassSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("purple"), 5, this);
    SimpleLineSymbol* fourthClassSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("purple"), 6, this);
    SimpleLineSymbol* fifthClassSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Solid, QColor("purple"), 7, this);

    // Create five class breaks.
    ClassBreak* firstClassBreak = new ClassBreak("Under 500", "0 - 500", 0.0, 500.0, firstClassSymbol, this);
    ClassBreak* secondClassBreak = new ClassBreak("501 to 1000", "501 - 1000", 501.0, 1000.0, secondClassSymbol, this);
    ClassBreak* thirdClassBreak = new ClassBreak("1001 to 1500", "1001 - 1500", 1001.0, 1500.0, thirdClassSymbol, this);
    ClassBreak* fourthClassBreak = new ClassBreak("1501 to 2000", "1501 - 2000", 1501.0, 2000.0, fourthClassSymbol, this);
    ClassBreak* fifthClassBreak = new ClassBreak("2001 to 2300", "2001 to 2300", 2001.0, 2300.0, fifthClassSymbol, this);
    QList<ClassBreak*> elevationBreaks = {firstClassBreak, secondClassBreak, thirdClassBreak, fourthClassBreak, fifthClassBreak};

    // Create and assign a class breaks renderer to the feature layer.
    ClassBreaksRenderer* elevationClassBreaksRenderer = new ClassBreaksRenderer("ELEV_GAIN", elevationBreaks, this);
    featureLayer->setRenderer(elevationClassBreaksRenderer);

    // Set the layer opacity to semi-transparent.
    featureLayer->setOpacity(0.75);
}

void Display_a_map::addBikeOnlyTrailsLayer()
{
    // Create a trails feature layer and add it to the map view.
    FeatureLayer* featureLayer = addFeatureLayer(trailsQUrl);

    // Write a definition expression to filter for trails that permit the use of bikes.
    featureLayer->setDefinitionExpression("USE_BIKE = 'Yes'");

    // Create and assign a simple renderer to the feature layer.
    SimpleLineSymbol* bikeTrailSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Dot, QColor("blue"), 2, this);
    SimpleRenderer* bikeTrailRenderer = new SimpleRenderer(bikeTrailSymbol, this);
    featureLayer->setRenderer(bikeTrailRenderer);
}

void Display_a_map::addNoBikeTrailsLayer()
{
    // Create a trails feature layer and add it to the map view.
    FeatureLayer* featureLayer = addFeatureLayer(trailsQUrl);

    // Write a definition expression to filter for trails that don't permit the use of bikes.
    featureLayer->setDefinitionExpression("USE_BIKE = 'No'");

    // Create and assign a simple renderer to the feature layer.
    SimpleLineSymbol* noBikeTrailSymbol = new SimpleLineSymbol(SimpleLineSymbolStyle::Dot, QColor("red"), 2, this);
    SimpleRenderer* noBikeTrailRenderer = new SimpleRenderer(noBikeTrailSymbol, this);
    featureLayer->setRenderer(noBikeTrailRenderer);
}

LabelDefinition* Display_a_map::makeLabelDefinition()
{
    // Create a new text symbol.
    TextSymbol* labelTextSymbol = new TextSymbol(this);
    labelTextSymbol->setColor(QColor("white"));
    labelTextSymbol->setSize(12.0);
    labelTextSymbol->setHaloColor (QColor("black"));
    labelTextSymbol->setHaloWidth(2.0);
    labelTextSymbol->setFontFamily("Arial");
    labelTextSymbol->setFontStyle(FontStyle::Italic);
    labelTextSymbol->setFontWeight(FontWeight::Normal);

    // Create a new Arcade label expression based on the field name: 'TRL_NAME'.
    ArcadeLabelExpression* labelExpression = new ArcadeLabelExpression("$feature.TRL_NAME", this);

    // Create and return the label definition.
    return new LabelDefinition(labelExpression, labelTextSymbol, this);
}

void Display_a_map::addTrailheadsLayer()
{
    // Create a ServiceFeatureTable from the URL, create a FeatureLayer, and add that to the map's operational layers.
    QUrl trailheadsQUrl("https://services3.arcgis.com/GVgbJbqm8hXASVYi/arcgis/rest/services/Trailheads/FeatureServer/0");
    FeatureLayer* featureLayer = addFeatureLayer(trailheadsQUrl);

    // Create a new picture marker symbol that uses the trailhead image.
    QUrl trailheadImageQUrl("https://static.arcgis.com/images/Symbols/NPS/npsPictograph_0231b.png");
    PictureMarkerSymbol* pictureMarkerSymbol = new PictureMarkerSymbol(trailheadImageQUrl, this);
    pictureMarkerSymbol->setHeight(18.0);
    pictureMarkerSymbol->setWidth(18.0);

    // Create a new simple renderer based on the picture marker symbol.
    SimpleRenderer* simpleRenderer = new SimpleRenderer(pictureMarkerSymbol, this);

    // Set the feature layer's renderer and enable labels.
    featureLayer->setRenderer(simpleRenderer);
    featureLayer->setLabelsEnabled(true);

    // Create the label definition.
    LabelDefinition* trailHeadsDefinition = makeLabelDefinition();

    // Add the label definition to the layer's label definition collection.
    featureLayer->labelDefinitions()->append(trailHeadsDefinition);
}

// Set the view (created in QML)
void Display_a_map::setMapView(MapQuickView* mapView)
{
    if (!mapView || mapView == m_mapView)
    {
        return;
    }

    m_mapView = mapView;
    m_mapView->setMap(m_map);

    setupViewpoint();

    emit mapViewChanged();
}

